/* -*- Mode: C; indent-tabs-mode: nil; c-basic-offset: 8-*- */

/*
 *This file is part of GNU MlView
 *
 *GNU MlView is free software; 
 *you can redistribute it and/or modify it under the terms of 
 *the GNU General Public License as 
 *published by the Free Software Foundation; either version 2, 
 *or (at your option) any later version.
 *
 *GNU MlView is distributed in the hope 
 *that it will be useful, but WITHOUT ANY WARRANTY; 
 *without even the implied warranty of 
 *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *See the GNU General Public License for more details.
 *
 *You should have received a copy of the 
 *GNU General Public License along with MlView; 
 *see the file COPYING. If not, write to 
 *the Free Software Foundation, Inc., 
 *59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *Copyright 2001-2002 Dodji Seketeli.
 */


#include <stdio.h>
#include <locale.h>
#include <ctype.h>
#include "mlview-utils.h"
#include <gnome.h>
#include "config.h"
#include <libxml/valid.h>

/**
 *@file
 *Misc usefull functions.
 *There are also some functions that could be small patches
 *gtk and libxml2.
 */

typedef struct _ErrorMessage ErrorMessage ;

struct _ErrorMessage{
  gchar * error_type ;
  gchar *error_string ;
};

static ErrorMessage p_error_messages_tab[]={
  {"error","OK"}, /*0*/
  {"error","parser internal error"},
  {"error","no more memory"},
  {"error","error at document start"}, /*3*/
  {"error","document empty"},
  {"error","document end reached unexpectedly"},
  {"error","invalid hexadecimal character reference"}, /*6*/
  {"error","invalid decimal character reference"},
  {"error","invalid chararacter reference"},
  {"error","invalid character"},
  {"error","character reference at end of file"}, /*10*/
  {"error","character reference in prolog"},
  {"error","character reference in epilog"}, /*12*/
  {"error","character reference in DTD"},
  {"error","entity reference at end of file"}, /*14*/
  {"error","entity reference in prolog"},
  {"error","entity reference in epilog"}, /*16*/
  {"error","entity reference in DTD"},
  {"error","pe reference at end of file"}, /*18*/
  {"error","pe reference in prolog"},
  {"error","pe reference in epilog"}, /*20*/
  {"error","pe reference in int subset"},
  {"error","entity reference with no name"}, /*22*/
  {"error","entity reference with semicolon missing"},
  {"error","pe reference with no name"}, /*24*/
  {"error","pe reference semicolon missing"},
  {"error","undeclared entity"}, /*26*/
  {"warning","undeclared entity"},
  {"error","unparsed entity"}, /*28*/
  {"error","entity is external"},
  {"error","entity is parameter"},/*30*/
  {"error","unknown encoding"},
  {"error","unsupported encoding"}, /*32*/
  {"error","string not started"},
  {"error","string not closed"}, /*34*/
  {"error","namespace declaration error"},
  {"error","entity not started"} , /*36*/
  {"error","entity not finished"},

  {"error","\'<\' character in attribute"}, /*38*/
  {"error","attribute not started"},
  {"error","entity not finished"}, /*40*/
  {"error","attribute without value"}, 
  {"error","attribute redefined"},/*42*/
  {"error","literal not started"},
  {"error","literal not finished"}, /*44*/
  {"error","comment not finished"},
  {"error","processing instruction not started"}, /*46*/
  {"error","processing instruction not finished"},
  {"error","notation not started"}, /*48*/
  {"error","notation not finished"},
  {"error","attribute list not star*ted"}, /*50*/
  {"error","attribute list not finished"},
  {"error","mixed content not started"}, /*52*/
  {"error","mixed content not finished"},
  {"error","element content not started"}, /*54*/
  {"error","element content not finished"},
  {"error","xml declaration not started"},/*56*/
  {"error","xml declaration not finished"},
  {"error","condition section not started"},/*58*/
  {"error","condition section not finished"},
  {"error","external subset not finished"},/*60*/
  {"error","doctype not finished"},
  {"error","misplaced CDATA end"},/*62*/
  {"error","CDATA not finished"},
  {"error","reserved xml name"},/*64*/
  {"error","space required"},
  {"error","separator required"}, /*66*/
  {"error","NM token required"},
  {"error","name token required"}, /*68*/
  {"error","PCDATA required"},
  {"error","URI required"},/*70*/
  {"error","public id required"},
  {"error","\'<\' required"}, /*72*/
  {"error","\'>\' required"},
  {"error","\'</\' required"},/*74*/
  {"error","\'=\' required"},
  {"error","tag name mismatch"}, /*76*/
  {"error","tag not finished"},
  {"error","standalone value"},/*78*/
  {"error","encoding name"},
  {"error","hyphen in comment"},/*80*/
  {"error","invalid encoding"},
  {"error","entity standalone"},/*82*/
  {"error","condition section invalid"},
  {"error","value required"},/*84*/
  {"error","not well balanced"},
  {"error","extra content"},/*86*/
  {"error","entity character error"},
  {"error","PE internal"},/*88*/
  {"error","entity loop"},
  {"error","entity boundary"},/*90*/
  {"error","invalid uri"},
  {"error","uri fragment"},/*92*/  
  {NULL}
} ;


/*The dynamic list of available encodings*/
static GList * gv_available_encodings = NULL ;
gint gv_available_encodings_ref_count = 0 ;


/*
 *The list of the default supported encodings
 */
guchar * gv_default_encodings[] = {
	"UTF-8", 
	"ISO_8859-1",
	"UTF-16",
	NULL/*must be the last item of the list!!*/
} ;

/**
 *Sets the decoration of the window so that
 *it has got a decoration homegenous with the
 *other MlView windows. Basically, it sets the
 *the MlView icon for example.
 *Note that the windows must be *NOT* realized before
 *this function is called.
 *
 *@param a_w the instance of GtkWindow to decorate.
 *@param a_ctxt the current instance of #MlViewAppContext
 *@param a_wm_class_name the class name of the window as it
 *should be seen by the window manager. (just give a custom string
 *that identifies the type of window e.g error-dialog or file-selection-dialog
 *
 */
void
mlview_utils_set_window_deco (GtkWindow * a_w,
                              MlViewAppContext *a_ctxt,
                              guchar * a_wm_class_name)
{
        g_return_if_fail (a_w && GTK_IS_WINDOW (a_w)) ;
        g_return_if_fail (a_ctxt && a_wm_class_name) ;
        
        if (GTK_WIDGET_REALIZED (GTK_WIDGET (a_w)) == TRUE)
                return ;

        gtk_window_set_wmclass (a_w, a_wm_class_name, "MlView") ;

        gtk_widget_realize (GTK_WIDGET (a_w)) ;

        mlview_app_context_set_window_icon (a_ctxt, GTK_WIDGET (a_w)) ;
}

/*
 *xmlParserErrors messages retrieval routines.
 */

/**
 *retrieves the error message associated to the error 
 *state of the current parser context.
 *@param a_xml_parser_context the parser context to consider
 *@param a_error_string the error message 
 *associated to the error state of the parser context. Is
 *set to null if no message if associated to the error 
 *state of the parser context. The caller
 *of this function must deallocated this variable.
 *@author Dodji Seketeli <sdodji@yahoo.com> 
 *
 */
void 
mlview_utils_get_xml_parser_error_as_string1 
(xmlParserCtxtPtr a_xml_parser_context, 
 gchar * a_error_string)
{
  ErrorMessage error_message ;
  g_return_if_fail(a_xml_parser_context!=NULL) ;

  a_error_string=NULL ;    
  
  if (a_xml_parser_context->errNo >= sizeof (p_error_messages_tab)-1)
    return ;
  error_message = p_error_messages_tab[a_xml_parser_context->errNo] ;
  
  a_error_string = 
	  g_malloc(strlen (error_message.error_type) + 
		   strlen (error_message.error_string) + 2) ;
  sprintf (a_error_string, 
	   "%s:%s",
	   error_message.error_type,
	   error_message.error_string) ;
}

/**
 *retrieves the error message associated 
 *to the error number given in argument.
 *@param a_xml_error the error number.
 *@param a_error_string the error 
 *message associated to the error state of the Parser context. Is
 *set to null if no message if associated to 
 *the error state of the parser context. The caller
 *of this function must deallocated this variable.
 *@author Dodji Seketeli <sdodji@yahoo.com> 
 *
 */
void 
mlview_utils_get_xml_parser_error_as_string2 (xmlParserErrors a_xml_error, 
					      gchar * a_error_string)
{
  ErrorMessage error_message ;
  a_error_string=NULL ;  
  
  if (a_xml_error >= sizeof (p_error_messages_tab) -1)
    return ;
  error_message = error_message = p_error_messages_tab[a_xml_error] ;

  a_error_string = 
	  g_malloc(strlen (error_message.error_type) +
		   strlen (error_message.error_string) + 2) ;
  sprintf (a_error_string, 
	   "%s:%s",
	   error_message.error_type,
	   error_message.error_string) ;
}

/**
 *Tests if a given string is a white string.
 *@param a_str the string to consider.
 *@return TRUE if a_str is a white string, FALSE if not.
 */
gboolean
mlview_utils_is_white_string (const gchar *a_str)
{
	gchar * tmp_str= (gchar *) a_str ;
	if (tmp_str == NULL)
		return FALSE ;	

	while (*tmp_str) {
		if (!isspace (*tmp_str)) return FALSE ;			
		tmp_str ++ ;
	}
	return TRUE ;
}


/**
 *Given an ISOLAT1 string, this function
 *returns the size (in bytes) this string
 *would have occupied if it was encoded in utf-8.
 *@param a_str the  NULL terminated string to consider.
 *@param a_result_len
 */
enum MlViewStatus
mlview_utils_isolat1_str_len_as_utf8 (guchar *a_str,
                                      gint *a_result_len)
{
	gint len = 0 ;
	guchar *char_ptr = NULL ;

	g_return_val_if_fail (a_str != NULL
			      && a_result_len != NULL,
			      MLVIEW_BAD_PARAM_ERROR) ;

	for (char_ptr = a_str ; 
	     *char_ptr ;
	     char_ptr ++) 
        {

		if (*char_ptr <= 0x7F) 
                {
			/*the utf-8 char would take 1 byte*/
			len += 1 ;
		} 
                else 
                {
			/*the utf-8 char would take 2 bytes*/
			len += 2 ;
		}
	}

	*a_result_len = len ;

	return MLVIEW_OK ;
}

/**
 *Compute the number of ISO 8859_1 characters contained
 *in a null terminated utf8 string.
 *@param a_utf8_str the utf8 string input to consider. Must be
 *not NULL.
 *@a_len out parameter the computed length. This parameter
 *is valid if and only if this function returns MLVIEW_OK. Must be
 *not NULL.
 *@return <ul>
 *<li>MLVIEW_OK upon successufull completion</li>
 *<li>MLVIEW_BAD_PARAM_ERROR if one the parameter is erratic.</li>
 *<li>MLVIEW_BAD_PARAM_ERROR if one the parameter is erratic.</li>
 *<li>MLVIEW_ENCODING_ERROR if a_utf8_str is not a well formed utf8 string</li>
 *<li>MLVIEW_CHAR_TOO_LONG if one of the characters contained
 *in a_utf8_str is too long to be an ISO 8859_1 character. In this
 *case, a_len points to the number of characters in a_utf8_str up to the
 *character that is too long (exclusive)</li>
 *</ul>
 */
enum MlViewStatus
mlview_utils_utf8_str_len_as_isolat1 (guchar *a_utf8_str,
				      gint *a_len)
{
	/*
	 *Note: this function can be made shorter
	 *but it considers all the cases of the utf8 encoding
	 *to ease further extensions ...
	 */

	guchar *byte_ptr = NULL ;
	gint len = 0 ;

	/*to store the final decoded 
	 *unicode char
	 */
	guint c = 0 ;

	g_return_val_if_fail (a_utf8_str && a_len,
			      MLVIEW_BAD_PARAM_ERROR) ;
	*a_len = 0 ;
	
	for (byte_ptr = a_utf8_str ;
	     byte_ptr && *byte_ptr ; 
		byte_ptr++) {

		gint nb_bytes_2_decode = 0 ;

		if (*byte_ptr <= 0x7F) {
			/*
			 *7 bits long char
			 *encoded over 1 byte:
			 * 0xxx xxxx
			 */
			c = *byte_ptr ;
			nb_bytes_2_decode = 1 ;

		} else if ((*byte_ptr & 0xE0) == 0xC0) {
			/*
			 *up to 11 bits long char.
			 *encoded over 2 bytes:
			 *110x xxxx  10xx xxxx
			 */
			c = *byte_ptr & 0x1F ;
			nb_bytes_2_decode = 2 ;

		} else if ((*byte_ptr & 0xF0) == 0xE0) {
			/*
			 *up to 16 bit long char
			 *encoded over 3 bytes:
			 *1110 xxxx  10xx xxxx  10xx xxxx
			 */
			c = *byte_ptr & 0x0F ;
			nb_bytes_2_decode = 3 ;

		} else if ((*byte_ptr & 0xF8) == 0xF0) {
			/*
			 *up to 21 bits long char
			 *encoded over 4 bytes:
			 *1111 0xxx  10xx xxxx  10xx xxxx  10xx xxxx
			 */
			c = *byte_ptr & 0x7 ;
			nb_bytes_2_decode = 4 ;

		} else if ((*byte_ptr & 0xFC) == 0xF8) {
			/*
			 *up to 26 bits long char
			 *encoded over 5 bytes.
			 *1111 10xx  10xx xxxx  10xx xxxx  
			 *10xx xxxx  10xx xxxx
			 */
			c = *byte_ptr & 3 ;
			nb_bytes_2_decode = 5 ;

		} else if ((*byte_ptr & 0xFE) == 0xFC) {
			/*
			 *up to 31 bits long char
			 *encoded over 6 bytes:
			 *1111 110x  10xx xxxx  10xx xxxx  
			 *10xx xxxx  10xx xxxx  10xx xxxx
			 */
			c = *byte_ptr & 1 ;
			nb_bytes_2_decode = 6 ;

		} else {
			/*BAD ENCODING*/
			return MLVIEW_ENCODING_ERROR ;
		}

		/*
		 *Go and decode the remaining byte(s)
		 *(if any) to get the current character.
		 */
		for ( ;
		      nb_bytes_2_decode > 1 ;
		      nb_bytes_2_decode --) {
			/*decode the next byte*/
			byte_ptr ++ ;

			/*byte pattern must be: 10xx xxxx*/
			if ((*byte_ptr & 0xC0) != 0x80) {
				return MLVIEW_ENCODING_ERROR ;
			}

			c = (c << 6) | (*byte_ptr & 0x3F) ;
		}

		if (c <= 0xFF) {/*Add other conditions to support
				 *other char sets (ucs2, ucs3, ucs4).
				 */
			len ++ ;
		} else {
			/*the char is too long to fit
			 *into the supposed charset len.
			 */
			return MLVIEW_CHAR_TOO_LONG ;
		}
	}

	*a_len = len ;

	return MLVIEW_OK ;
}


/**
 *Translates a_in_str in a string encoded in utf8.
 *For the time being, a_str is supposed to 
 *be encoded in an isolat1 charset.
 *@param a_in_str the string to translate.
 *@param a_out_str out parameter. a pointer
 *to an allocated buffer where to store the translated string.
 *Note that the caller deallocate *a_out_str 
 *after completion of this function.
 *
 *@return 
 *<ul>
 *<li>MLVIEW_OK if a_str has been successfully translated.
 *<li> error codes in case of error.
 *</ul>
 */
enum MlViewStatus
mlview_utils_isolat1_str_to_utf8 (guchar *a_in_str,
				  guchar **a_out_str)
{
	enum MlViewStatus status = MLVIEW_OK ;
	gint len1 = 0, len2 = 0 ;

	g_return_val_if_fail (a_in_str != NULL
			      && a_out_str != NULL,
			      MLVIEW_BAD_PARAM_ERROR) ;

	*a_out_str = NULL ;
	len1 = strlen (a_in_str) ;

	if (len1 == 0)
		return MLVIEW_OK ;
	
	status =
		mlview_utils_isolat1_str_len_as_utf8 (a_in_str,
						      &len2) ;
	if (status != MLVIEW_OK)
		return status ;
	
	if (len2 == 0)
		return MLVIEW_OK ;

	*a_out_str = g_malloc0 (len2 + 1) ;

	if (isolat1ToUTF8 (*a_out_str, &len2,
			   a_in_str, &len1)) 
        {

		g_free (*a_out_str) ;
		*a_out_str = NULL ;
                return MLVIEW_ERROR ;
	}

	return MLVIEW_OK ;
}


/**
 *Converts an utf8 string to an ucs1 string.
 *@param a_in_str the NULL terminated 
 *string to convert.
 *@param a_out_str out param. 
 *A pointer to the converted string.
 *Must be free by the caller.
 *@return MLVIEW_OK upon successfull completion,
 *
 *
 */
enum MlViewStatus
mlview_utils_utf8_str_to_isolat1 (guchar *a_in_str,
				  guchar **a_out_str)
{
	enum MlViewStatus status = MLVIEW_OK ;
	gint len1 = 0, len2 = 0 ;

	g_return_val_if_fail (a_in_str && a_out_str,
			      MLVIEW_BAD_PARAM_ERROR) ;

	*a_out_str = NULL ;

	len2 = strlen (a_in_str) ;

	if (len2 == 0)
		return MLVIEW_OK ;

	status = mlview_utils_utf8_str_len_as_isolat1 (a_in_str,
						       &len1) ;

	g_return_val_if_fail (status == MLVIEW_OK, status) ;

	*a_out_str = g_malloc0 (len1 + 1) ;

	if (UTF8Toisolat1 (*a_out_str, &len1,
			   a_in_str, &len2)) {

		/*function failed*/
		g_free (*a_out_str) ;
		*a_out_str = NULL ;
		status = MLVIEW_ENCODING_ERROR ;

	} else {
		
		/*function returned successfully*/
		status = MLVIEW_OK ;
	}

	return status ;
}


/**
 *Initialises the gettext i18n functions.
 */
void
mlview_utils_init_i18n (void)
{
	setlocale (LC_ALL, "") ;
	bindtextdomain (PACKAGE, PACKAGE_LOCALE_DIR) ;
	textdomain (PACKAGE) ;
}

/**
 *This method increments the reference count
 *of the "available encodings list".
 *See mlview_utils_editor_unref_available_encodings()
 *to decrement the reference count and free the
 *"available encodings list" variable.
 *
 *To avoid memory leaks, 
 *Every single object that uses the
 *"available encodings list" variable should call
 *this method once at its creation and call
 *mlview_utils_unref_available_encodings()
 *at its destruction.
 */
void
mlview_utils_ref_available_encodings (void)
{
	if (gv_available_encodings)
		gv_available_encodings_ref_count ++ ;
}


/**
 *This method decrements the reference count of
 *the "available encoding list" global variable. 
 *If the reference count of that variable goes
 *down to zero, this method destroys the variable.
 *See mlview_utils_ref_available_encodings() for
 *more explanation about how these methods should be used.
 */
void
mlview_utils_unref_available_encodings (void)
{
	if (gv_available_encodings == NULL)
		return ;

	if (gv_available_encodings_ref_count)
		gv_available_encodings_ref_count -- ;

	if (gv_available_encodings_ref_count == 0) {
		/*free the gv_available_encodings list*/
		GList * list_elem = NULL ;

		for (list_elem = gv_available_encodings ;
		     list_elem ;
		     list_elem = list_elem->next) {
			
			if (list_elem->data) {
				g_free (list_elem->data) ;
				list_elem->data = NULL ;
			}
		}

		g_list_free (gv_available_encodings) ;
		gv_available_encodings = NULL ;		
	}
}


/**
 *Initialyses the list of the available encodings
 *supported by mlview.
 *This method must be called at least once to make
 *the least available.
 *
 */
void
mlview_utils_init_available_encodings_list (void)
{
	if (gv_available_encodings == NULL) {
		guchar *enc = NULL ;
		gint i = 0 ;

		/*
		 *initialise the libxml2 encoding support, 
		 *in case it hasn't been initialized before ...
		 */
		xmlInitCharEncodingHandlers () ;

		/*initialise the list*/
		for (i=0, enc = gv_default_encodings[i] ;
		     enc ;
		     i++, enc = gv_default_encodings[i]) {

			gv_available_encodings =
				g_list_append (gv_available_encodings,
					       g_strdup (enc)) ;
		}
	}
}


/**
 *Gets the list of the available
 *encodings supported by mlview.
 *
 *The returned list must *NOT* be
 *freed by the caller.
 *It should intead use the functions
 *mlview_utils_ref_available_encodings()
 *and mlview_utils_unref_available_encodings()
 *to control the lifecyle of the list.
 */
GList *
mlview_utils_get_available_encodings (void)
{	
	if (gv_available_encodings == NULL) {
		mlview_utils_init_available_encodings_list () ;
	}

	return gv_available_encodings ;
}


/**
 *An implementation of the 
 *glib GCompareFunc.
 *Compares two strings.
 *
 *@param a_str1 the of the two strings to compare.
 *@param a_str2 the second of the two strings to compare.
 *@return TRUE if the two strings are equal, FALSE otherwise.
 */
static gboolean
string_compare (const guchar * a_str1,
		const guchar * a_str2) 
{
	if (a_str1 == NULL || a_str2 == NULL)
		return FALSE ;

	if (!strcmp (a_str1, a_str2))
		return TRUE ;

	return FALSE ;
}


static void
mlview_utils_free_encoding_handler_if_needed 
(xmlCharEncodingHandler * a_enc_handler)
{
	if (a_enc_handler == NULL)
		return ;

	if (a_enc_handler->iconv_in) {

		iconv_close (a_enc_handler->iconv_in) ;
		a_enc_handler->iconv_in = NULL ;
	} else {
		return ;
	}

	if (a_enc_handler->iconv_out) {
		
		iconv_close (a_enc_handler->iconv_out) ;
		a_enc_handler->iconv_out = NULL ;
	} else {
		return ;
	}

	if (a_enc_handler->name) {
		xmlFree (a_enc_handler->name) ;
		a_enc_handler->name = NULL ;
	}

	xmlFree (a_enc_handler) ;
	return ;
}

/**
 *Tests wether a given encoding is supported by mlview 
 *(actually, the underlying libxml2) or not.
 *
 *@param a_encoding the encoding name preferably 
 *styled as the iconv encoding naming style.
 *On GNU/Linux systems, type "iconv --list" to have 
 *an idea about the iconv encoding naming style.
 *
 *@return TRUE if the encoding is supported FALSE if otherwise.
 */
gboolean
mlview_utils_is_encoding_supported (guchar *a_encoding)
{
	xmlCharEncodingHandler * enc_handler = NULL ;

	
	if (a_encoding == NULL)
		return FALSE ;

	enc_handler =
		xmlFindCharEncodingHandler (a_encoding) ;

	if (!enc_handler)
		return FALSE ;

	mlview_utils_free_encoding_handler_if_needed (enc_handler) ;

	return TRUE ;
}


/**
 *Adds an ecoding to the "available list of encodings"
 *supported by mlview. 
 *This method actually checks if the new encoding
 *is supported by the underlying libxml parser.
 *If yes, it adds the encoding to the list of available
 *encodings if and only if the list does not contain it
 *already.
 *
 *@param a_name the name of the encoding. Must be 
 *respect the naming scheme of the GNU/Linux version of iconv.
 *
 *@return 
 *<ul>
 *<li>MLVIEW_OK upon successfull completion, 
 *MLVIEW_NO_ENCODINGS if the list of available encodings supported
 *by mlview has not been built prior to calling this method.
 *In that case, call mlview_utils_get_available_encodings()
 *to build it.</li>
 *<li>
 *MLVIEW_ENCODING_NOT_SUPPORTED if the encoding is not
 *supported by the underlying libxml2 parser.
 *</li>
 *</ul>
 */
enum MlViewStatus
mlview_utils_add_supported_encoding (guchar *a_name)
{
	if (gv_available_encodings == NULL)
		return MLVIEW_NO_ENCODINGS_ERROR ;

	/*make sure libxml2 supports this encoding*/
	if (mlview_utils_is_encoding_supported (a_name) == FALSE)
		return MLVIEW_ENCODING_NOT_SUPPORTED_ERROR ;

	/*
	 *make sure the encoding is not already in the list
	 *and add it.
	 */
	if (!g_list_find_custom (gv_available_encodings, a_name,
				 (GCompareFunc) string_compare)) {

		gv_available_encodings =
			g_list_append (gv_available_encodings,
				       g_strdup (a_name)) ;
	}

	return MLVIEW_OK ;
}


/*
 *FIXME: define a macro to free
 *a char handler handler if and only
 *if it must be freed (the caller)
 */

enum MlViewStatus
mlview_utils_del_supported_encoding (guchar *a_name)
{
	GList *list_elem = NULL ;

	if (gv_available_encodings == NULL)
		return MLVIEW_NO_ENCODINGS_ERROR ;

	if ( (list_elem = g_list_find_custom (gv_available_encodings,
					    a_name,
					    (GCompareFunc) string_compare)) ) {
		gv_available_encodings =
			g_list_remove_link (gv_available_encodings,
					    list_elem) ;

		if (list_elem && list_elem->data) {
			g_free (list_elem->data) ;
			g_list_free (list_elem) ;
		}
	}

	return MLVIEW_OK ;
}

/*=============================================
 * some usefull patches to libxml:
 *
 * xmlUnlinkNs (xmlNode * node, xmlNs * ns)
 * xmlUnlinkNsDef (xmlNode * node, xmlNs * ns)
 *============================================*/

/**
 *Unlinks a namespace from an xmlNode.
 *@param node the node to unlink the namespace
 *from.
 *@param ns the namespace to unlink.
 *@return the unlinked namespace, or NULL
 *if the functions failed or could not unlink
 *the namespace.
 */
xmlNs *
xmlUnlinkNs (xmlNode * node, xmlNs * ns)
{
	xmlNs * cur ;
	
	if(node == NULL  ||  ns == NULL) {
		return NULL;
	}

	cur = node->ns ;

	if(cur == NULL)
		return NULL ;

	/*
	 *Check if the namespace to unlink is 
	 *the first ns in the xml node nsDef list. If yes, unlink it
	 **/
	if(ns == cur){

		if (ns->next){

			cur = ns->next ;
			ns->next = NULL ;
		}

		return ns ;
	}

	/*
	 *the namespace to unlink may be 
	 *further in the ns list. 
	 *Walk to the namespace just before it...
	 */
	while(cur != NULL  &&  cur->next != ns) {

		cur = cur->next ;
	}

	if(cur == NULL)
		return NULL ;

	/*
	 *at this stage, cur points to the 
	 *namespace that is just before the 
	 *namespace to unlink. Let's now unlink ns
	 */
	if(ns->next == NULL) {

		cur->next = NULL ;

	} else {
		cur->next = ns->next ;
		ns->next = NULL ;
	}

	return ns ;
	
}


/**
 *Unlinks a namespace definition
 *from an xmlNode .
 *@param node the xml node to unlink the namespace
 *definition from.
 *@param ns the namespace to unlink.
 */
xmlNs *
xmlUnlinkNsDef (xmlNode * node, xmlNs * ns)
{
	xmlNs * cur ;

	if(node == NULL  ||  ns == NULL) {
		return NULL;
	}
	
	cur = node->nsDef ;
	if(cur == NULL)
		return NULL ;

	/*
	 *Check if the namespace to unlink is the 
	 *first ns in the xml node 
	 *nsDef list. If yes, unlink it
	 */
	if(ns == cur){
		node->nsDef = node->nsDef->next ;
		ns->next = NULL ;
		return ns ;
	}

	/*
	 *the namespace to unlink may be 
	 *further in the ns list. 
	 *Walk to the namespace just before it...
	 */
	while(cur != NULL  &&  cur->next != ns) {
		cur = cur->next ;
	}
	if(cur == NULL)
		return NULL ;

	/*
	 *at this stage, cur points to the 
	 *namespace that is just before the 
	 *namespace to unlink. 
	 *Let's now unlink ns
	 */
	if(ns->next == NULL)
		cur->next = NULL ;
	else{
		cur->next = ns->next ;
		ns->next = NULL ;
	}
	return ns ;
}


/**
 *Gets the valid children of the element a_node.
 *This function replaces xmlValidGetValidElement because it
 *works even if a_node is the 
 *root node and has no children... I've remarked that
 *xmlValidGetValidElements () does not work in that case :( 
 *
 *@param a_list the output element name array.
 *Must be allocated by the user. 
 *@param a_max the max size of the output list.
 *@param a_node the node to consider.
 *
 *
 */
int
xmlValidGetValidElementsChildren (xmlNode *a_node, 
				  const xmlChar **a_list,
				  int a_max)
{
	xmlElement * element_desc = NULL ;
	gint nb_valid_elements = 0 ;

	/*some sanity checks...*/
	if (a_node == NULL) return -2 ;
	if (a_list == NULL) return -2 ;
	if (a_list == NULL)  return -2 ;
	if (a_max == 0) return -2 ;
	if (a_node->type != XML_ELEMENT_NODE) return -2 ;
	if (a_node->parent == NULL) return -2 ;

	if (a_node->children) {
		/*the node is the root node 
		 *and it has children. We can it valid children
		 *with the function
		 */
		return xmlValidGetValidElements (a_node->last, NULL, 
						 a_list, a_max) ;
	
	} else {
		gint nb_elements = 0, i = 0 ;
		const xmlChar * elements[256] ;
		xmlNode * test_node = NULL ;
		xmlValidCtxt vctxt ;
		
		memset(&vctxt, 0, sizeof (xmlValidCtxt));

		element_desc = 
			xmlGetDtdElementDesc (a_node->parent->doc->intSubset,
					      a_node->name) ;

		if ((element_desc == NULL) 
		    && (a_node->parent->doc->extSubset != NULL)) {
			element_desc = 
				xmlGetDtdElementDesc
				(a_node->parent->doc->extSubset,
				 a_node->name) ;
		}

		if (element_desc == NULL) return (-1) ;
		
		/*
		 *Create a dummy child element
		 */
		test_node = 
			xmlNewChild (a_node, NULL, "<!dummy?>", NULL) ;

                if (test_node->name)
                {
                        xmlFree ((xmlChar *)test_node->name) ;
                        test_node->name = NULL ;
                }

		/*
		 *Insert each potential child node 
		 *and check if the current node is still valid.
		 */
		nb_elements = 
			xmlValidGetPotentialChildren (element_desc->content,
						      elements, 
						      &nb_elements, 256) ;
		
		for (i = 0 ; i < nb_elements ; i++) {
			test_node->name = elements[i] ;
			if (xmlStrEqual(test_node->name, "#PCDATA"))
				test_node->type = XML_TEXT_NODE ;
			else
				test_node->type = XML_ELEMENT_NODE ;

			if (xmlValidateOneElement (&vctxt, 
						   a_node->parent->doc, 
						   a_node)) {
				int j = 0 ;
				for (j = 0 ; j < nb_valid_elements ; j++)
					if (xmlStrEqual (elements[i],
							 a_list[j])) {
						break ;
					}

				a_list[nb_valid_elements++] = elements[i] ;
				
				if (nb_valid_elements >= a_max) break ;
			}
		}

		/*
		 *Restore the tree structure
		 */
		xmlUnlinkNode (test_node) ;

                if (test_node)
                {
                        test_node->name = NULL ;
                        xmlFreeNode (test_node) ;
                        test_node = NULL ;
                }

		return nb_valid_elements ;
	}
}

/**
 *Parses a string to get a node name and a
 *namespace out of it.
 *
 *@param a_node the xml node that is 
 *supposed to hold the fully qualified name.
 *This parameter is usefull to access the information
 *about namespace definition store in the document
 *holding this node if any.
 *
 *@param a_full_name the 
 *name to parse to extract 
 *the prefix part and the local name part.
 *
 *@param a_ns out parameter. 
 *The parsed namespace.
 *Can be null if there is no prefix or 
 *if the prefix has not been declared in 
 *the node or in one of the node's parents.
 *NOTE THAT THE USER MUST *NOT* free this pointer !!!.
 *
 *@param a_local_name out parameter. 
 *The local name part of the full name. 
 *Can be null if there is no local name part.
 *If a_local_name is not null, 
 *the caller has to free it ; 
 */
void
mlview_utils_parse_full_name (xmlNode *a_node, 
			      guchar *a_full_name, 
			      xmlNs ** a_ns, 
			      guchar **a_local_name)
{
	guchar *colon_ptr = NULL, *prefix = NULL, *local_name = NULL ;
	gchar **full_name_array = NULL ;
	xmlNs * ns = NULL ;

	g_return_if_fail (a_node != NULL) ;
	g_return_if_fail (a_full_name != NULL) ;

	*a_ns = NULL ;
	*a_local_name = NULL ;

	colon_ptr = index (a_full_name, ':') ;

	if (colon_ptr != NULL) 
        {
		full_name_array = g_strsplit (a_full_name,":", 2) ;
		prefix = full_name_array[0] ;
		local_name = full_name_array[1] ;
		ns = xmlSearchNs (a_node->doc,
				  a_node,
				  prefix) ;
		if(ns != NULL)
			*a_ns = ns ;
		else
			*a_ns = NULL ;

		if (local_name != NULL 
		    && ! mlview_utils_is_white_string (local_name))
			*a_local_name = g_strstrip (local_name) ;
		
	} 
        else 
        {
		*a_local_name = 
			g_strdup (g_strstrip (a_full_name)) ;
	}
}


/**
 *Displays an error message in a error dialog graphically.
 *The error dialog box is modal.
 *
 *@param a_msg_format the format string of the
 *message, just the same as in the C standard lib
 *printf first parameter.
 *@param ... a list of argument just like the list of
 *argument of the C standard lib printf list of
 *arguments.
 */
void
mlview_utils_display_error_dialog (MlViewAppContext * a_ctxt,
                                   gchar *a_msg_format, ...)
{
        va_list params ;        
        GnomeDialog * err_dialog = NULL ;
        GnomeApp * app = NULL ;
        GtkText * text_widget = NULL ;
        gchar * err_msg = NULL ;
        
        g_return_if_fail (a_msg_format != NULL) ;

        va_start (params, a_msg_format) ;
        
        err_msg = g_strdup_vprintf (a_msg_format, params) ;

        va_end (params) ;
                        
        g_return_if_fail (err_msg != NULL) ;

        app = mlview_app_context_get_element (a_ctxt, "GnomeApp") ;

        g_return_if_fail (app && GTK_IS_WINDOW (app)) ;

        text_widget = GTK_TEXT (gtk_text_new (NULL, NULL)) ;

        gtk_text_set_editable (text_widget, FALSE) ;

        gtk_text_set_line_wrap (text_widget, TRUE) ;

        gtk_text_set_point (text_widget, 0) ;

        gtk_text_insert (text_widget, NULL, NULL, NULL, 
                         err_msg, strlen (err_msg)) ;
        

        err_dialog = GNOME_DIALOG (gnome_dialog_new (_("Error"),
                                                     GNOME_STOCK_BUTTON_OK,
                                                     NULL)) ;

        mlview_utils_set_window_deco (GTK_WINDOW (err_dialog),
                                      a_ctxt, "error-dialog") ;

        g_return_if_fail (err_dialog) ;

        gnome_dialog_close_hides (err_dialog, FALSE) ;

        gtk_box_pack_start_defaults (GTK_BOX (err_dialog->vbox),
                                     GTK_WIDGET (text_widget)) ;

        /*the error dialog stays on top of the mlview app*/
        gnome_dialog_set_parent (err_dialog, GTK_WINDOW (app)) ;

        /*let the window be user resizable*/
        gtk_window_set_policy (GTK_WINDOW (err_dialog),
                               FALSE, TRUE, FALSE) ;

        gtk_widget_show_all (GTK_WIDGET (err_dialog)) ;

        gnome_dialog_run (err_dialog) ;

        gtk_widget_destroy (GTK_WIDGET (err_dialog)) ;

        if (err_msg)
        {
                g_free (err_msg) ;
                err_msg = NULL ;
        }
}

/*=============================================
  some usefull patches i wrote for the files gtkclist.c and gtkctree.c:

  gtk_clist_absolute_row_top_ypixel(GtkCList *a_clist, gint a_row)
  gtk_ctree_node_absolute_top_ypixel(GtkCTree * a_tree, GtkCTreeNode *a_node)
 *============================================*/

#define CELL_SPACING 1

gint
gtk_clist_absolute_row_top_ypixel (GtkCList *a_clist, gint a_row)
{
	g_return_val_if_fail(a_clist != NULL,-1) ;
	g_return_val_if_fail(GTK_IS_CLIST(a_clist), -1) ;

	return (a_clist->row_height * a_row  +
		(a_row + 1) * CELL_SPACING) ;
}

gint
gtk_ctree_node_absolute_top_ypixel (GtkCTree * a_tree, GtkCTreeNode *a_node)
{
	gint pos ;
	g_return_val_if_fail (a_tree != NULL, -1) ;
	g_return_val_if_fail (GTK_IS_CTREE(a_tree), -1) ;
	g_return_val_if_fail (a_node != NULL, -1) ;

	pos = g_list_position (a_tree->clist.row_list,&a_node->list) ;
	return gtk_clist_absolute_row_top_ypixel (&a_tree->clist, pos) ;
}
